/*
 * ff2dirac_dirac.c - Utilities common to Dirac and Schroedinger codecs.
 * Copyright (C) 2010 BBC, Anuradha Suraparaju<asuraparaju@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with This program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifdef HAVE_KATE
#include "kate/kate.h"
#endif
#include "libavformat/avformat.h"
#include "ogg/ogg.h"
#include "ffmpeg2dirac.h"
#include "oggmuxutils.h"
#include "ff2dirac_dirac.h"
#include <assert.h>

#define F2D_ARRAY_SIZE(x) (sizeof(x)/sizeof(x[0]))

typedef struct
{
    int width;
    int height;
    int frame_rate_num;
    int frame_rate_denom;
} ff2dirac_videoformat_info_t;

static const ff2dirac_videoformat_info_t video_format_info[] = {
    { 640,  480,  24000, 1001},
    { 176,  120,  15000, 1001},
    { 176,  144,  25,    2   },
    { 352,  240,  15000, 1001},
    { 352,  288,  25,    2   },
    { 704,  480,  15000, 1001},
    { 704,  576,  25,    2   },
    { 720,  480,  30000, 1001},
    { 720,  576,  25,    1   },
    { 1280, 720,  60000, 1001},
    { 1280, 720,  50,    1   },
    { 1920, 1080, 30000, 1001},
    { 1920, 1080, 25,    1   },
    { 1920, 1080, 60000, 1001},
    { 1920, 1080, 50,    1   },
    { 2048, 1080, 24,    1   },
    { 4096, 2160, 24,    1   },
    { 3840, 2160, 60000, 1001},
    { 3840, 2160, 50,    1   },
    { 7680, 4320, 60000, 1001},
    { 7680, 4320, 50,    1   },
};

extern int ff2dirac_get_video_format_idx(ff2dirac ctx)
{
    unsigned int ret_idx = 0;
    unsigned int idx;
    unsigned int num_formats = sizeof(video_format_info) /
                               sizeof(video_format_info[0]);

    for (idx = 1 ; idx < num_formats; ++idx ) {
        const ff2dirac_videoformat_info_t *vf =
                                        &video_format_info[idx];
        if (ctx->frame_width  == vf->width &&
            ctx->frame_height == vf->height){
            ret_idx = idx;
            if (ctx->framerate.num == vf->frame_rate_num &&
                ctx->framerate.num == vf->frame_rate_denom) {
                return idx;
            }
        }
    }
    return ret_idx;
}

extern void ff2dirac_calculate_granulepos (ff2dirac_privdata_t *dd, ogg_packet *op)
{
    ogg_int64_t granulepos_hi;
    ogg_int64_t granulepos_low;
    int pt, dt, dist, delay;
    if (dd->is_sync_point)
        dd->distance_from_sync = 0;
    else
        dd->distance_from_sync++;

    pt = dd->presentation_frame_number * 2;
    dt = dd->decode_frame_number * 2;
    delay = pt - dt;
    dist = dd->distance_from_sync;

    granulepos_hi = ((pt - delay)<<9) | ((dist>>8));
    granulepos_low = (delay << 9) | (dist & 0xff);

    op->granulepos = (granulepos_hi << 22) | (granulepos_low);
    op->packetno = dd->packet_no++;
    dd->decode_frame_number++;
}

/* Granule shift is always 22 for Dirac */
static const int DIRAC_GRANULE_SHIFT = 22;

extern double ff2dirac_granule_time(ff2dirac_video_enc_state_t *dd, ogg_int64_t granulepos)
{
    ogg_int64_t pt;
    ff2dirac_privdata_t *cinfo = dd->vcodec->priv_data;

    if (granulepos == -1)
        return -1;

    pt = ((granulepos >> DIRAC_GRANULE_SHIFT) + (granulepos & ((1ULL<<DIRAC_GRANULE_SHIFT)-1)))>>9;

    pt >>= 1;

    return pt * ((double)cinfo->fps_denom/cinfo->fps_num);
}

extern int ff2dirac_granule_shift(ff2dirac_video_enc_state_t *dd)
{
    return DIRAC_GRANULE_SHIFT;
}

static char *mvprec_list[] = { "1", "1/2", "1/4", "1/8" };
extern int ff2dirac_get_mv_precision_idx(char *mv_prec)
{
    int i, num_mvprec = F2D_ARRAY_SIZE(mvprec_list);

    for (i = 0; i < num_mvprec; ++i)
    {
        if (strcmp(mvprec_list[i], mv_prec) == 0)
            return i;
    }
    fprintf(stderr, "Supported mv precision values are\n");
    for (i = 0; i < num_mvprec; ++i)
        fprintf(stderr, "\t%s\n", mvprec_list[i]);
    return -1;
}

static char *wavelet_list[] = {
  "desl_dubuc_9_7",
  "le_gall_5_3",
  "desl_dubuc_13_7",
  "haar_0",
  "haar_1",
  "fidelity",
  "daub_9_7"
};

extern int ff2dirac_get_wavelet_idx(char *wavelet)
{
    int i, num_wavelets = F2D_ARRAY_SIZE(wavelet_list);

    for (i = 0; i < num_wavelets; ++i)
    {
        if (strcmp(wavelet_list[i], wavelet) == 0)
            return i;
    }
    fprintf(stderr, "Wavelet filter %s not supported. Supported wavelet filters are\n", wavelet);
    for (i = 0; i < num_wavelets; ++i)
        fprintf(stderr, "\t%s\n", wavelet_list[i]);
    return -1;
}

extern ogg_int64_t ff2dirac_get_duration(ff2dirac_video_enc_state_t *dd)
{
    ff2dirac_privdata_t *cinfo = dd->vcodec->priv_data;
    ogg_int64_t duration = (cinfo->fps_denom * 1000) / cinfo->fps_num;

    return duration;
}

extern int ff2dirac_packet_iskeyframe(ogg_packet *op)
{
    unsigned int prefix = 0;
    assert(op->bytes >= 5);
    prefix = (op->packet[0] << 24) +
            (op->packet[1] << 16) +
            (op->packet[2] << 8)  +
            op->packet[3];
    assert(prefix ==  0x42424344);

    /* start of a sequence is considered a key frame*/
    return op->packet[4] == 0x00;
}
