/* -*- tab-width:4;c-file-style:"cc-mode"; -*- */
/*
 * libdirac_encoder_utils.c - Encoder module that uses dirac-research codec
 * Copyright (C) 2010      BBC, Anuradha Suraparaju <asuraparaju@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with This program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifdef HAVE_DIRAC
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>

#include "ogg/ogg.h"
#include "libavutil/common.h"
#include "libavformat/avformat.h"
#include "libdirac_encoder/dirac_encoder.h"
#include "oggmuxutils.h"
#include "ffmpeg2dirac.h"
#include "ff2dirac_dirac.h"

#ifndef DIRAC_RESEARCH_VERSION_ATLEAST
# define DIRAC_RESEARCH_VERSION_ATLEAST(x,y,z) 0
#endif

static char *prefilter_list[] = {"none", "diaglp", "rectlp", "cwm"};

static int get_prefilter_idx(char *prefilter)
{
    int i;
    int num_prefilters = F2D_ARRAY_SIZE(prefilter_list);

    for (i = 0; i < num_prefilters; ++i)
    {
        if (strcmp(prefilter_list[i], prefilter) == 0)
            return i;
    }
    fprintf(stderr, "Prefilter %s not supported. Dirac Encoder could not be set up\n", prefilter);
    fprintf(stderr, "Supported pre-filters for Dirac are:\n");
    for (i = 0; i < num_prefilters; ++i)
        fprintf(stderr, "\t%s\n", prefilter_list[i]);
    return -1;
}

static int ff2libdirac_encode_header(ff2dirac_video_enc_state_t *dd, ogg_packet *op)
{
    ff2dirac_privdata_t *cinfo = dd->vcodec->priv_data;
    uint8_t *header = cinfo->dirac_buf;

    if (header[0] != 'B' ||
        header[1] != 'B' ||
        header[2] != 'C' ||
        header[3] != 'D' ||
        header[4] != 0x0)
    {
        fprintf (stderr, "Not a sequence header\n");
        return -1;
    }
    long header_len = (header[5] << 24) +
                      (header[6] << 16) +
                      (header[7] << 8) +
                      header[8];

    if (header_len <= 13)
    {
        fprintf (stderr, "Sequence header length <=13. Too small\n");
        return -1;
    }
    if (header_len > cinfo->dirac_buf_size)
    {
        fprintf (stderr, "Sequence header length greater than packet length\n");
        return -1;
    }
    op->b_o_s = 1;
    op->e_o_s = 0;
    op->bytes = header_len;
    op->packet = header;
    op->granulepos = 0;
    op->packetno = 0;

    return 0;
}

static int ff2libdirac_encode_clear (ff2dirac_video_enc_state_t *dd)
{
    ff2dirac_privdata_t *cinfo = dd->vcodec->priv_data;
    dirac_encoder_close(dd->di);
    av_free(cinfo->dirac_buf);
    av_free(cinfo->dirac_frame_buf);
    av_free(cinfo);
    dd->di = NULL;
    dd->vcodec->priv_data = NULL;
    return 0;
}

static int ff2libdirac_load_frame(ff2dirac_video_enc_state_t *dd, AVPicture *pic)
{
    ff2dirac_privdata_t *cinfo = dd->vcodec->priv_data;
    dirac_encoder_t *enc = dd->di;
    dirac_sourceparams_t *src_params = &((dirac_encoder_context_t *)cinfo->ctx)->src_params;

    avpicture_layout(pic, cinfo->ff_pix_fmt,
                     src_params->width, src_params->height,
                     cinfo->dirac_frame_buf, cinfo->dirac_frame_size);

    dirac_encoder_load(enc, cinfo->dirac_frame_buf, cinfo->dirac_frame_size);

    return 0;
}

static int ff2libdirac_encode_packetout(ff2dirac_video_enc_state_t *dd, int e_o_s, ogg_packet *op)
{
    dirac_encoder_t *di = dd->di;
    ff2dirac_privdata_t *cinfo = dd->vcodec->priv_data;

    int ret = 0;
    int go = 1;
    if (e_o_s && cinfo->eos_signalled == 0)
    {
        cinfo->eos_signalled = 1;
        dirac_encoder_end_sequence(di);
    }

    if (cinfo->eos_pulled)
        go = 0;

    if (go)
    {
        dirac_encoder_state_t state;
        di->enc_buf.buffer = cinfo->dirac_buf;
        di->enc_buf.size = cinfo->dirac_buf_size;

        state = dirac_encoder_output(di);
        switch(state)
        {
        case ENC_STATE_BUFFER:
            ret = 0;
            break;
        case ENC_STATE_AVAIL:
        case ENC_STATE_EOS:
            assert (di->enc_buf.size > 0);
            op->b_o_s = 0;
            if (di->enc_buf.buffer[4] == 0x00)
            {
                cinfo->is_sync_point = 1;
            }
            else
            {
                cinfo->is_sync_point = 0;
            }
            op->e_o_s = 0;
            op->bytes = di->enc_buf.size;
            op->packet = di->enc_buf.buffer;
            if (di->enc_pparams.pnum >= 0)
            {
                cinfo->presentation_frame_number = di->enc_pparams.pnum;
                ff2dirac_calculate_granulepos(dd->vcodec->priv_data, op);
            }
            if (state == ENC_STATE_EOS)
            {
                op->e_o_s = 1;
                cinfo->eos_pulled = 1;
            }
            ret = 1;
            break;
        case ENC_STATE_INVALID:
            fprintf(stderr, "Irrecoverable error in Dirac encoder library\n");
            exit(1);
            break;
        default:
            fprintf (stderr, "Should not reach here. Unknown state....\n");
            exit(1);
        }
    }
    return ret;
}

static int ff2libdirac_encode_init (ff2dirac ctx, ff2dirac_video_enc_state_t *dd)
{
    ff2dirac_privdata_t *priv_data = av_mallocz(sizeof(ff2dirac_privdata_t));
    ogg_packet op_hdr;

    dirac_encoder_context_t enc_ctx;

    priv_data->preset = ff2dirac_get_video_format_idx(ctx);
    dirac_encoder_context_init(&enc_ctx, priv_data->preset);

    priv_data->ff_pix_fmt = ctx->pix_fmt;
    switch(ctx->pix_fmt)
    {
    case PIX_FMT_YUV420P:
        enc_ctx.src_params.chroma = format420;
        priv_data->dirac_buf_size = ctx->frame_width*ctx->frame_height*3/2;
        break;
    case PIX_FMT_YUV422P:
        enc_ctx.src_params.chroma = format422;
        priv_data->dirac_buf_size = ctx->frame_width*ctx->frame_height*2;
        break;
    case PIX_FMT_YUV444P:
        priv_data->dirac_buf_size = ctx->frame_width*ctx->frame_height*3;
        enc_ctx.src_params.chroma = format444;
        break;
    default:
        fprintf (stderr, "Unsupported pix format %d\n", ctx->pix_fmt);
        return -1;
    }
    enc_ctx.src_params.width = ctx->frame_width;
    enc_ctx.src_params.height = ctx->frame_height;
    if (ctx->frame_width*ctx->frame_height < (702*480*2)/3)
    {
        /* For small pictures, set better (more expensive) parameters */
        enc_ctx.enc_params.mv_precision = MV_PRECISION_QUARTER_PIXEL;
        enc_ctx.enc_params.xblen = 12;
        enc_ctx.enc_params.yblen = 12;
        enc_ctx.enc_params.xbsep = 8;
        enc_ctx.enc_params.ybsep = 8;
    }

    enc_ctx.src_params.frame_rate.numerator = ctx->framerate.num;
    enc_ctx.src_params.frame_rate.denominator = ctx->framerate.den;

    if (ctx->aspect_numerator && ctx->aspect_denominator)
    {
        enc_ctx.src_params.pix_asr.numerator = ctx->aspect_numerator;
        enc_ctx.src_params.pix_asr.denominator = ctx->aspect_denominator;
    }

    enc_ctx.enc_params.trate = ctx->video_bitrate/1000;

    enc_ctx.enc_params.qf = ctx->video_quality;

    enc_ctx.enc_params.lossless = ctx->lossless;

    if (ctx->numL1 >= 0)
        enc_ctx.enc_params.num_L1 = ctx->numL1;
    if (ctx->sepL1 >= 0)
        enc_ctx.enc_params.L1_sep = ctx->sepL1;

    if (ctx->prefilter)
    {
        if (ctx->prefilter_strength == -1)
        {
            fprintf (stderr,
                 "prefilter-strength not specified. The Dirac encoder could not set up\n");
            return -1;
        }
        enc_ctx.enc_params.prefilter_strength = ctx->prefilter_strength;

        enc_ctx.enc_params.prefilter = get_prefilter_idx(ctx->prefilter);
        if (enc_ctx.enc_params.prefilter == -1)
            return -1;
    }

    if (ctx->mv_prec)
    {
        enc_ctx.enc_params.mv_precision = ff2dirac_get_mv_precision_idx(ctx->mv_prec);
        if (enc_ctx.enc_params.mv_precision == -1)
        {
            fprintf (stderr, "Motion vector precision %s not supported. The Dirac encoder could not set up\n", ctx->mv_prec);
            return -1;
        }
    }

    /* xbspep, ybsep, xblen and yblen should be multiples of 4 */
    if (ctx->mc_block_width > -1 )
    {
        enc_ctx.enc_params.xbsep = ctx->mc_block_width/4 * 4;
        if ( enc_ctx.enc_params.xbsep != ctx->mc_block_width)
            fprintf (stderr, "Motion block width modified to be multiple of 4\n");
    }
    if (ctx->mc_block_height > -1 )
    {
        enc_ctx.enc_params.ybsep = ctx->mc_block_height/4 * 4;
        if (enc_ctx.enc_params.ybsep != ctx->mc_block_height)
            fprintf (stderr, "Motion block height modified to be multiples of 4\n");
    }

    if (ctx->mc_block_overlap > -1)
    {
        enc_ctx.enc_params.xblen = enc_ctx.enc_params.xbsep * (100 + ctx->mc_block_overlap)/400 * 4;
        enc_ctx.enc_params.yblen = enc_ctx.enc_params.ybsep * (100 + ctx->mc_block_overlap)/400 * 4;
    }

#if DIRAC_RESEARCH_VERSION_ATLEAST(1, 0, 1)
    enc_ctx.enc_params.combined_me = ctx->combined_me;
#else
    fprintf(stderr, "combined_me option supprted only in Dirac version 1.0.1 or later\n");
#endif
    if (ctx->dwt_intra)
    {
        enc_ctx.enc_params.intra_wlt_filter = ff2dirac_get_wavelet_idx(ctx->dwt_intra);
        if (enc_ctx.enc_params.intra_wlt_filter == -1)
        {
            fprintf (stderr, "Intra wavelet filter in the Dirac encoder could not be set up\n");
            return -1;
        }
    }

    if (ctx->dwt_inter)
    {
        enc_ctx.enc_params.inter_wlt_filter = ff2dirac_get_wavelet_idx(ctx->dwt_inter);
        if (enc_ctx.enc_params.inter_wlt_filter == -1)
        {
            fprintf (stderr, "Inter wavelet filter in the Dirac encoder could not be set up\n");
            return -1;
        }
    }

    if (ctx->dwt_depth > -1)
        enc_ctx.enc_params.wlt_depth = ctx->dwt_depth;
    enc_ctx.enc_params.multi_quants = ctx->multi_quants;
    if (ctx->no_spartition)
        enc_ctx.enc_params.spatial_partition = 0;

    if (ctx->cpd > -1.0f)
        enc_ctx.enc_params.cpd = ctx->cpd;

    /* Arithmetic coding can be disabled only for Intra only sequences */
    if (ctx->no_ac && enc_ctx.enc_params.num_L1 != 0)
    {
        fprintf (stderr,
             "--no-ac option available for intra only sequences.\n"
             "The Dirac encoder could not set up.\n");
        return -1;
    }
    enc_ctx.enc_params.using_ac = !ctx->no_ac;

    enc_ctx.enc_params.full_search = ctx->full_search;
    dd->di = dirac_encoder_init(&enc_ctx, ctx->verbose);
    if (!dd->di){
        fprintf (stderr,
             "The Dirac encoder could not set up a mode according to\n"
             "the requested quality or bitrate.\n\n");
        return -1;
    }

    /* Allocate for the encoded output buffer */
    priv_data->dirac_buf = av_malloc (priv_data->dirac_buf_size);
    if (!priv_data->dirac_buf)
    {
        dirac_encoder_close(dd->di);
        fprintf (stderr,
             "Could not allocate %d bytes for encoder output buffer.\n"
             "The Dirac encoder could not set up\n", priv_data->dirac_buf_size);
        return -1;
    }

    /* Allocate for the frame buffer */
    switch (enc_ctx.src_params.chroma)
    {
    case format420:
        priv_data->dirac_frame_size = enc_ctx.src_params.width*enc_ctx.src_params.height*3/2;
        break;
    case format422:
        priv_data->dirac_frame_size = enc_ctx.src_params.width*enc_ctx.src_params.height*2;
        break;
    case format444:
        priv_data->dirac_frame_size = enc_ctx.src_params.width*enc_ctx.src_params.height*3;
        break;
    default:
        priv_data->dirac_frame_size = 0;
        break;
    }

    if (priv_data->dirac_frame_size == 0 || !(priv_data->dirac_frame_buf = av_mallocz (priv_data->dirac_frame_size)))
    {
        av_free(priv_data->dirac_buf);
        dirac_encoder_close(dd->di);
        return -1;
    }
    priv_data->fps_num = enc_ctx.src_params.frame_rate.numerator;
    priv_data->fps_denom = enc_ctx.src_params.frame_rate.denominator;
    dd->vcodec->priv_data = priv_data;
    priv_data->ctx = &((dirac_encoder_t *)dd->di)->enc_ctx;

    /* encode dummy frames so that we can generate a b_o_s page */
    do
    {
        AVPicture pic;
        avpicture_fill(&pic, priv_data->dirac_frame_buf, ctx->pix_fmt,
                       ctx->frame_width, ctx->frame_height);
        ff2libdirac_load_frame(dd, &pic);
    } while (ff2libdirac_encode_packetout(dd, 0, &op_hdr) == 0);

    dirac_encoder_close(dd->di);
    dd->di = dirac_encoder_init(&enc_ctx, ctx->verbose);
    priv_data->ctx = &((dirac_encoder_t *)dd->di)->enc_ctx;
    priv_data->decode_frame_number = -1;
    priv_data->presentation_frame_number = 0;
    priv_data->distance_from_sync = 0;
    priv_data->packet_no = 0;
    dd->vcodec->priv_data = priv_data;
    return 0;
}


ff2dirac_vcodec_t ff2libdirac_encoder = {
  VCODEC_DIRAC,
  NULL,
  NULL,
  ff2libdirac_encode_init,
  ff2libdirac_encode_header,
  ff2libdirac_load_frame,
  ff2libdirac_encode_packetout,
  ff2libdirac_encode_clear,
  ff2dirac_granule_time,
  ff2dirac_granule_shift,
  NULL,
  NULL,
  ff2dirac_get_duration,
  ff2dirac_packet_iskeyframe
};
#endif
