/* -*- tab-width:4;c-file-style:"cc-mode"; -*- */
/*
 * libschroedinger_encoder_utils.c - Encoder modules that uses Schroedinger
 * Copyright (C) 2010 BBC, Anuradha Suraparaju <asuraparaju@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with This program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifdef HAVE_SCHROEDINGER
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>

#include "ogg/ogg.h"
#include "libavutil/common.h"
#include "libavformat/avformat.h"
#include "schroedinger/schro.h"
#include "oggmuxutils.h"
#include "ffmpeg2dirac.h"
#include "ff2dirac_dirac.h"
#include "libschroedinger_options.h"

static int64_t g_pts = 0;

static int ff2libschroedinger_encode_header(ff2dirac_video_enc_state_t *dd, ogg_packet *op)
{
    ff2dirac_privdata_t *cinfo = dd->vcodec->priv_data;
    uint8_t *header = cinfo->dirac_buf;

    if (header[0] != 'B' ||
        header[1] != 'B' ||
        header[2] != 'C' ||
        header[3] != 'D' ||
        header[4] != 0x0)
    {
        fprintf (stderr, "Not a sequence header\n");
        return -1;
    }
    long header_len = (header[5] << 24) +
                      (header[6] << 16) +
                      (header[7] << 8) +
                      header[8];

    if (header_len <= 13)
    {
        fprintf (stderr, "Sequence header length <=13. Too small\n");
        return -1;
    }
    if (header_len > cinfo->dirac_buf_size)
    {
        fprintf (stderr, "Sequence header length greater than packet length\n");
        return -1;
    }
    op->b_o_s = 1;
    op->e_o_s = 0;
    op->bytes = header_len;
    op->packet = header;
    op->granulepos = 0;
    op->packetno = 0;

    return 0;
}

static int ff2libschroedinger_encode_clear (ff2dirac_video_enc_state_t *dd)
{
    ff2dirac_privdata_t *cinfo = dd->vcodec->priv_data;
    schro_encoder_free(dd->di);
    libschroedinger_clear_options();
    av_free(cinfo->ctx);
    av_free(cinfo->dirac_buf);
    av_free(cinfo);
    dd->di = NULL;
    dd->vcodec->priv_data = NULL;
    return 0;
}

static void free_schro_frame(SchroFrame *frame, void *priv)
{
    AVPicture *p_pic = priv;

    if (!p_pic)
        return;

    avpicture_free(p_pic);
    av_freep(&p_pic);
}

static SchroFrame *create_schro_frame(ff2dirac_privdata_t *cinfo, SchroFrameFormat ff)
{
    SchroVideoFormat *format = cinfo->ctx;
    SchroFrame *p_frame;
    int y_width, uv_width;
    int y_height, uv_height;
    int i;
    AVPicture *p_pic;

    y_width = format->width;
    y_height = format->height;
    uv_width = y_width >> (SCHRO_FRAME_FORMAT_H_SHIFT(ff));
    uv_height = y_height >> (SCHRO_FRAME_FORMAT_V_SHIFT(ff));

    p_pic = av_mallocz(sizeof(AVPicture));
    avpicture_alloc(p_pic, cinfo->ff_pix_fmt, y_width, y_height);

    p_frame = schro_frame_new();
    p_frame->format = ff;
    p_frame->width = y_width;
    p_frame->height = y_height;
    schro_frame_set_free_callback(p_frame, free_schro_frame, (void *)p_pic);

    for (i = 0; i < 3; ++i) {
        p_frame->components[i].width  = i ? uv_width : y_width;
        p_frame->components[i].stride = p_pic->linesize[i];
        p_frame->components[i].height = i ? uv_height : y_height;
        p_frame->components[i].length =
                 p_frame->components[i].stride * p_frame->components[i].height;
        p_frame->components[i].data   = p_pic->data[i];

        if (i) {
            p_frame->components[i].v_shift =
                SCHRO_FRAME_FORMAT_V_SHIFT(p_frame->format);
            p_frame->components[i].h_shift =
                SCHRO_FRAME_FORMAT_H_SHIFT(p_frame->format);
        }
    }

    return p_frame;
}

static int ff2libschroedinger_load_frame(ff2dirac_video_enc_state_t *dd, AVPicture *pic)
{
    SchroEncoder *enc = dd->di;
    ff2dirac_privdata_t *cinfo = (ff2dirac_privdata_t *)dd->vcodec->priv_data;
    SchroVideoFormat *format = cinfo->ctx;
    SchroFrameFormat ff;
    SchroFrame *frame;

    int64_t *p_pts = av_malloc(sizeof(int64_t));

    *p_pts = g_pts++;

    switch (format->chroma_format)
    {
    case SCHRO_CHROMA_420:
        ff = SCHRO_FRAME_FORMAT_U8_420;
        break;
    case SCHRO_CHROMA_422:
        ff = SCHRO_FRAME_FORMAT_U8_422;
        break;
    case SCHRO_CHROMA_444:
        ff = SCHRO_FRAME_FORMAT_U8_444;
        break;
    default:
        fprintf (stderr, "Unsupported Schro chroma format %d\n", format->chroma_format);
        return -1;
        break;
    }

    if (!(frame = create_schro_frame(cinfo, ff)))
    {
        fprintf (stderr, "Could not allocate schro frame\n");
        return -1;
    }

    if (frame) {
        avpicture_layout(pic, cinfo->ff_pix_fmt,
                         format->width, format->height,
                         frame->components[0].data, cinfo->dirac_frame_size);
    }

    schro_encoder_push_frame_full(enc, frame, p_pts);

    return 0;
}

static int ff2libschroedinger_encode_packetout(ff2dirac_video_enc_state_t *dd, int e_o_s, ogg_packet *op)
{
    SchroEncoder *di = dd->di;
    ff2dirac_privdata_t *cinfo = dd->vcodec->priv_data;

    int ret = 0;
    int go = 1;
    if (e_o_s && cinfo->eos_signalled == 0)
    {
        cinfo->eos_signalled = 1;
        schro_encoder_end_of_stream(di);
    }

    if (cinfo->eos_pulled)
        go = 0;

    if (go)
    {
        SchroStateEnum state;
        int presentation_frame;
        SchroBuffer *enc_buf;
        int64_t *p_pts;
        void *vptr = NULL;

        state = schro_encoder_wait(di);
        switch(state)
        {
        case SCHRO_STATE_NEED_FRAME:
            ret = 0;
            break;
        case SCHRO_STATE_END_OF_STREAM:
            cinfo->eos_pulled = 1;
            ret = 0;
            break;
        case SCHRO_STATE_HAVE_BUFFER:
            enc_buf = schro_encoder_pull_full(di, &presentation_frame, &vptr);
            p_pts = vptr;
            op->b_o_s = 0;
            if (SCHRO_PARSE_CODE_IS_SEQ_HEADER(enc_buf->data[4]))
            {
                cinfo->is_sync_point = 1;
            }
            else
            {
                cinfo->is_sync_point = 0;
            }
            op->e_o_s = 0;
            memcpy (cinfo->dirac_buf, enc_buf->data, enc_buf->length);
            op->bytes = enc_buf->length;
            op->packet = cinfo->dirac_buf;

            /***
            //else if (SCHRO_PARSE_CODE_IS_PICTURE(enc_buf->data[4]))
            {
                cinfo->presentation_frame_number = (enc_buf->data[13] << 24) +
                                          (enc_buf->data[14] << 16) +
                                          (enc_buf->data[15] << 8) +
                                          enc_buf->data[16];
            }
            ***/
            if (p_pts != NULL && !SCHRO_PARSE_CODE_IS_END_OF_SEQUENCE(enc_buf->data[4]))
            {
                cinfo->presentation_frame_number = *p_pts;
                ff2dirac_calculate_granulepos(cinfo, op);
                av_free(p_pts);
            }
            else
            {
                op->e_o_s = 1;
            }
            schro_buffer_unref(enc_buf);
            ret = 1;
            break;
        case SCHRO_STATE_AGAIN:
            break;
        default:
            fprintf (stderr, "Should not reach here. Unknown state....\n");
            exit(1);
        }
    }
    return ret;
}

static char *prefilter_list[] = {"none", "centre_weighted_median", "gaussian", "add_noise", "adaptive_gaussian"};

static int get_prefilter_idx(char *prefilter)
{
    int i;
    int num_prefilters = F2D_ARRAY_SIZE(prefilter_list);

    for (i = 0; i < num_prefilters; ++i)
    {
        if (strcmp(prefilter_list[i], prefilter) == 0)
            return i;
    }
    fprintf(stderr, "Prefilter %s not supported. Schro Encoder could not be set up\n", prefilter);
    fprintf(stderr, "Supported pre-filters for Schroedinger are:\n");
    for (i = 0; i < num_prefilters; ++i)
        fprintf(stderr, "\t%s\n", prefilter_list[i]);
    return -1;
}

static int set_encoder_settings (ff2dirac ctx, ff2dirac_video_enc_state_t *dd)
{
    SchroEncoder *enc = dd->di;
    int optnum = 0;

    const LibSchroOptValuePair* const optlist = libschroedinger_get_encoder_options();

    /* The parameters below are common to all video codecs. So we have to use them if they have not
       been over-ridden by Schro specific options */
    if (ctx->video_bitrate > 0)
    {
        schro_encoder_setting_set_double ( enc, "rate_control",
                               SCHRO_ENCODER_RATE_CONTROL_CONSTANT_BITRATE);

        schro_encoder_setting_set_double (enc, "bitrate",
                                          ctx->video_bitrate);

    }
    else
    {
        schro_encoder_setting_set_double ( enc, "rate_control",
                        SCHRO_ENCODER_RATE_CONTROL_CONSTANT_QUALITY);

        schro_encoder_setting_set_double (enc, "quality",
                                          ctx->video_quality);
    }

    /* No schro specific options */
    if (optlist == NULL)
        return 0;

    while (optlist[optnum].idx != -1) {
        const SchroEncoderSetting *s = schro_encoder_get_setting_info(optlist[optnum].idx);
        double val;
        int i;

        switch (s->type) {
        case SCHRO_ENCODER_SETTING_TYPE_BOOLEAN:
            if ((strcmp(optlist[optnum].value , "true")) == 0)
                val = 1;
            else
                val = 0;
            schro_encoder_setting_set_double(enc, s->name, val);
            break;
        case SCHRO_ENCODER_SETTING_TYPE_INT:
            val = atoi(optlist[optnum].value);
            schro_encoder_setting_set_double(enc, s->name, val);
            break;
        case SCHRO_ENCODER_SETTING_TYPE_DOUBLE:
            val = atof(optlist[optnum].value);
            schro_encoder_setting_set_double(enc, s->name, val);
            break;
        case SCHRO_ENCODER_SETTING_TYPE_ENUM:
            for (i = 0; i <= s->max; ++i) {
                if (strcmp(optlist[optnum].value, s->enum_list[i]) == 0) {
                    schro_encoder_setting_set_double(enc, s->name, i);
                    break;
                }
            }
            break;
        case SCHRO_ENCODER_SETTING_TYPE_LAST:
            break;
        }

        ++optnum;
    }
    return 0;

}


static int ff2libschroedinger_encode_init (ff2dirac ctx, ff2dirac_video_enc_state_t *dd)
{
    ff2dirac_privdata_t *priv_data = av_mallocz(sizeof(ff2dirac_privdata_t));
    ogg_packet op_hdr;
    SchroVideoFormat *format = NULL;
    unsigned char *dummy_frame;

    schro_init();
    if (ctx->verbose)
        schro_debug_set_level(SCHRO_LEVEL_DEBUG);

    dd->di = schro_encoder_new();
    schro_encoder_set_packet_assembly(dd->di, 1);
    if (!dd->di){
        fprintf (stderr,
             "The Schro encoder could not set up a mode according to\n"
             "the requested quality or bitrate.\n\n");
        return -1;
    }

    format = schro_encoder_get_video_format(dd->di);
    priv_data->preset = ff2dirac_get_video_format_idx(ctx);
    schro_video_format_set_std_video_format(format, priv_data->preset);

    priv_data->ff_pix_fmt = ctx->pix_fmt;
    priv_data->dirac_buf_size = priv_data->dirac_frame_size = avpicture_get_size(ctx->pix_fmt, ctx->frame_width, ctx->frame_height);

    switch(ctx->pix_fmt)
    {
    case PIX_FMT_YUV420P:
        format->chroma_format = SCHRO_CHROMA_420;;
        break;
    case PIX_FMT_YUV422P:
        format->chroma_format = SCHRO_CHROMA_422;;
        break;
    case PIX_FMT_YUV444P:
        format->chroma_format = SCHRO_CHROMA_444;;
        break;
    default:
        fprintf (stderr, "Unsupported pix format %d\n", ctx->pix_fmt);
        priv_data->dirac_frame_size = 0;
        return -1;
    }
    format->width = ctx->frame_width;
    format->height = ctx->frame_height;
    format->frame_rate_numerator = ctx->framerate.num;
    format->frame_rate_denominator = ctx->framerate.den;
    priv_data->fps_num = format->frame_rate_numerator;
    priv_data->fps_denom = format->frame_rate_denominator;
    if (ctx->aspect_numerator && ctx->aspect_denominator)
    {
        format->aspect_ratio_numerator = ctx->aspect_numerator;
        format->aspect_ratio_denominator = ctx->aspect_denominator;
    }


    /* Allocate for the encoded output buffer */
    priv_data->dirac_buf = av_malloc (priv_data->dirac_buf_size);
    if (!priv_data->dirac_buf)
    {
        schro_encoder_free(dd->di);
        av_free(format);
        fprintf (stderr,
                "Could not allocate %d bytes for encoder output buffer.\n"
                "The Schro encoder could not be set up.\n",
                priv_data->dirac_buf_size);
        return -1;
    }

    /* Allocate for the frame buffer */
    if (priv_data->dirac_frame_size <= 0)
    {
        schro_encoder_free(dd->di);
        av_free(format);
        av_free(priv_data->dirac_buf);
        fprintf (stderr,
                "Invalid frame buffer size %d.\n"
                "The Schro encoder could not be set up.\n",
                priv_data->dirac_frame_size);
        return -1;
    }

    /* FIXME: Signal range hardcoded to 8-bit data until both libschroedinger
     * and libdirac support other bit-depth data. */
    schro_video_format_set_std_signal_range(format,
                                            SCHRO_SIGNAL_RANGE_8BIT_VIDEO);
    priv_data->ctx = format;
    schro_encoder_set_video_format(dd->di, format);
    dd->vcodec->priv_data = priv_data;

    if(set_encoder_settings(ctx, dd) == -1)
    {
        return -1;
    };

    schro_encoder_start(dd->di);

    /* encode dummy frames so that we can generate a b_o_s page */
    if ((dummy_frame = av_mallocz(priv_data->dirac_frame_size)) == NULL)
        return -1;
    do
    {
        AVPicture pic;
        avpicture_fill (&pic, dummy_frame, ctx->pix_fmt,
                        ctx->frame_width, ctx->frame_height);

        if (ff2libschroedinger_load_frame(dd, &pic) == -1)
        {
            schro_encoder_free(dd->di);
            av_free(format);
            av_free(priv_data->dirac_buf);
            av_free(dummy_frame);
            return -1;
        }
    } while (ff2libschroedinger_encode_packetout(dd, 0, &op_hdr) == 0);
    av_free(dummy_frame);

    schro_encoder_free(dd->di);
    dd->di = schro_encoder_new();
    schro_encoder_set_packet_assembly(dd->di, 1);
    schro_encoder_set_video_format(dd->di, format);
    set_encoder_settings(ctx, dd);
    schro_encoder_start(dd->di);

    priv_data->decode_frame_number = -1;
    priv_data->presentation_frame_number = 0;
    priv_data->distance_from_sync = 0;
    priv_data->packet_no = 0;
    g_pts = 0;
    return 0;
}


ff2dirac_vcodec_t ff2libschroedinger_encoder = {
  VCODEC_SCHRO,
  NULL,
  NULL,
  ff2libschroedinger_encode_init,
  ff2libschroedinger_encode_header,
  ff2libschroedinger_load_frame,
  ff2libschroedinger_encode_packetout,
  ff2libschroedinger_encode_clear,
  ff2dirac_granule_time,
  ff2dirac_granule_shift,
  NULL,
  NULL,
  ff2dirac_get_duration,
  ff2dirac_packet_iskeyframe
};
#endif /*HAVE_SCHROEDINGER*/
