/* -*- tab-width:4;c-file-style:"cc-mode"; -*- */
/*
 * libschroedinger_options.c - Utilities to process Schroedinger's encoder options dynamically
 * Copyright (C) 2010      BBC, Anuradha Suraparaju <asuraparaju@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with This program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifdef HAVE_SCHROEDINGER

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <schroedinger/schroencoder.h>

#include "libschroedinger_options.h"

int add_new_option (const SchroEncoderSetting *s, int * const flagptr, int flagval, struct option *newopt)
{
    assert(s != NULL);
    assert(flagptr != NULL);
    assert(newopt != NULL);

    char *optname = malloc(sizeof(char)*(strlen(s->name)+1));

    if (optname == NULL) {
        return -1;
    }
    strcpy(optname, s->name);

    newopt->name = optname;
    newopt->has_arg = required_argument;
    newopt->flag = flagptr;
    newopt->val = flagval;
    return 0;
}

int libschroedinger_add_encoding_options(struct option **optlist, int numopts, int * const flagptr, int flagval) {
    int dyn_opts = schro_encoder_get_n_settings();
    struct option *dynoptlist;
    int i;

    assert(*optlist!=NULL);
    assert(flagptr);

    dynoptlist = realloc(*optlist, sizeof(struct option)*(numopts+dyn_opts));
    if (dynoptlist == NULL) {
        /*realloc failed*/
        fprintf(stderr, "realloc failed! Could not allocate for Schroedinger's dynamic options\n");
        return -1;
    }

    *optlist = dynoptlist;

    for (i=0; i < dyn_opts; ++i) {
        const SchroEncoderSetting *s = schro_encoder_get_setting_info(i);
        if (add_new_option(s, flagptr, flagval+i, &dynoptlist[numopts+i]) == -1) {
            fprintf(stderr, "Could not add a schroedinger encoder option\n");
            return -1;
        }
    }
    return numopts+dyn_opts;
}

void libschroedinger_print_usage(void) {
    int num_opts = schro_encoder_get_n_settings();
    int i, j;
    fprintf (stdout, "\n    Schroedinger specific options:\n");
    for( i = 0; i < num_opts; ++i ) {
        const SchroEncoderSetting *s = schro_encoder_get_setting_info(i);

        switch(s->type) {
        case SCHRO_ENCODER_SETTING_TYPE_BOOLEAN:
            fprintf    (stdout, "      --%-30s Type bool      [%s]\n", s->name, s->default_value ? "true":"false");
            break;
        case SCHRO_ENCODER_SETTING_TYPE_INT:
            fprintf    (stdout, "      --%-30s Type int       [%d]    {%d..%d}\n", s->name, (int)(s->default_value), (int)(s->min), (int)(s->max));
            break;
        case SCHRO_ENCODER_SETTING_TYPE_DOUBLE:
            fprintf    (stdout, "      --%-30s Type double    [%g]    {%g..%g}\n", s->name, s->default_value, s->min, s->max);
            break;
        case SCHRO_ENCODER_SETTING_TYPE_ENUM:
            fprintf    (stdout, "      --%-30s Type enum      [%s]    {", s->name, s->enum_list[(int)(s->default_value)]);
            for (j=0; j<=s->max; ++j)
                fprintf (stdout, "%s ", s->enum_list[j]);

            fprintf (stdout, "}\n");
        default:
            break;
        }
    }
}

static LibSchroOptValuePair *schroOptionList = NULL;
static int optnum = 0;

int libschroedinger_set_encoder_option(int optidx, const char *value)
{
    const SchroEncoderSetting *s = schro_encoder_get_setting_info(optidx);
    int ival, j;
    double fval;

    if (schroOptionList == NULL) {
        schroOptionList = (LibSchroOptValuePair *)malloc(sizeof(LibSchroOptValuePair)*(schro_encoder_get_n_settings()+1));
        if (schroOptionList == NULL) {
            fprintf (stderr, "Could not allocate memory to store Schroedinger specific options\n");
            return -1;
        }
    }

    /* Testing if values are valid */
    switch (s->type) {
    case SCHRO_ENCODER_SETTING_TYPE_BOOLEAN:
        if (strcmp(value, "true") != 0 && strcmp(value, "false") != 0) {
            fprintf(stderr, "Allowable values for ard --%s are true or false\n", s->name);
            return -1;
        }
        break;
    case SCHRO_ENCODER_SETTING_TYPE_INT:
        ival = atoi(value);
        if (ival < s->min || ival > s->max) {
            fprintf(stderr, "Invalid value %s. Allowable range for arg --%s is {%d..%d}\n", value, s->name, (int)s->min, (int)s->max);
            return -1;
        }
    case SCHRO_ENCODER_SETTING_TYPE_DOUBLE:
        fval = atof(value);
        if (fval < s->min || fval > s->max) {
            fprintf(stderr, "Invalid value %s. Allowable range for arg --%s is {%g..%g}\n", value, s->name, s->min, s->max);
            return -1;
        }
        break;
    case SCHRO_ENCODER_SETTING_TYPE_ENUM:
        ival = -1;
        for (j = 0; j <= s->max; ++j ) {
            if (strcmp(value, s->enum_list[j]) == 0) {
                ival = j;
                break;
            }
        }
        if (ival == -1) {
            fprintf(stderr, "Invalid value %s for arg --%s. Valid values are {", value, s->name);
            for (j = 0; j <= s->max; ++j )
                fprintf(stderr, "%s ", s->enum_list[j]);
            fprintf(stderr, "}\n");
            return -1;
        }
        break;
    case SCHRO_ENCODER_SETTING_TYPE_LAST:
        break;
    }

    /* Finished validating the argument. Now store it for later use */
    schroOptionList[optnum].idx = optidx;

    schroOptionList[optnum].value = (char *)malloc(sizeof(char) * (strlen(value)+1));
    strcpy(schroOptionList[optnum].value, value);

    ++optnum;
    return 0;
}

const LibSchroOptValuePair *const libschroedinger_get_encoder_options(void)
{
    if (schroOptionList != NULL) {
        /* null terminate the list */
        schroOptionList[optnum].idx = -1;
        schroOptionList[optnum].value = NULL;
    }
    return schroOptionList;
}

void libschroedinger_clear_options(void) {
    int i = 0;

    if (schroOptionList != NULL) {
        while (schroOptionList[i].idx != -1 ) {
            free(schroOptionList[i].value);
            ++i;
        }
        free(schroOptionList);
       }
}

#endif /*HAVE_SCHROEDINGER*/
