/* -*- tab-width:4;c-file-style:"cc-mode"; -*- */
/*
 * libtheora_utils.c - Encoder module that uses libtheora
 * Copyright (C) 2010  BBC, Anuradha Suraparaju <asuraparaju@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with This program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifdef HAVE_THEORA
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>

#include "ogg/ogg.h"
#include "theora/theoraenc.h"
#include "libavutil/common.h"
#include "libavformat/avformat.h"
#include "oggmuxutils.h"
#include "ffmpeg2dirac.h"
#include "ff2dirac_theora.h"
#include "index.h"

static int ilog(unsigned _v){
  int ret;
  for(ret=0;_v;ret++)_v>>=1;
  return ret;
}

static int ff2libtheora_encode_header(ff2dirac_video_enc_state_t *dd, ogg_packet *op)
{
    ff2theora_privdata_t *info = dd->vcodec->priv_data;
    if(th_encode_flushheader(dd->di, &info->tc, op) <= 0) {
        fprintf(stderr, "Internal Theora library error.\n");
        exit(1);
    }
    return 0;
}

static int ff2libtheora_encode_clear (ff2dirac_video_enc_state_t *dd)
{
    oggmux_info *oinfo = dd->info;
    ff2theora_privdata_t *info = dd->vcodec->priv_data;
    if (oinfo->passno!=1)
        th_comment_clear(&info->tc);
    th_info_clear(&info->ti);
    th_encode_free(dd->di);
    dd->di = NULL;
    return 0;
}

static int ff2libtheora_load_frame(ff2dirac_video_enc_state_t *dd, AVPicture *pic)
{
    th_enc_ctx *td = dd->di;
    ff2theora_privdata_t *info= dd->vcodec->priv_data;
    static th_ycbcr_buffer ycbcr;
    oggmux_info *oinfo = dd->info;
    int ret;

    if(oinfo->passno==2){
        for(;;){
          static unsigned char buffer[80];
          static int buf_pos;
          int bytes;
          /*Ask the encoder how many bytes it would like.*/
          bytes=th_encode_ctl(td,TH_ENCCTL_2PASS_IN,NULL,0);
          if(bytes<0){
            fprintf(stderr,"Error submitting pass data in second pass.\n");
            exit(1);
          }
          /*If it's got enough, stop.*/
          if(bytes==0)break;
          /*Read in some more bytes, if necessary.*/
          if(bytes>80-buf_pos)bytes=80-buf_pos;
          if(bytes>0&&fread(buffer+buf_pos,1,bytes,oinfo->twopass_file)<bytes){
            fprintf(stderr,"Could not read frame data from two-pass data file!\n");
            exit(1);
          }
          /*And pass them off.*/
          ret=th_encode_ctl(td,TH_ENCCTL_2PASS_IN,buffer,bytes);
          if(ret<0){
            fprintf(stderr,"Error submitting pass data in second pass.\n");
            exit(1);
          }
          /*If the encoder consumed the whole buffer, reset it.*/
          if(ret>=bytes)buf_pos=0;
          /*Otherwise remember how much it used.*/
          else buf_pos+=ret;
        }
    }

    ycbcr[0].width = info->ti.frame_width;
    ycbcr[0].height = info->ti.frame_height;
    ycbcr[0].stride = pic->linesize[0];
    ycbcr[0].data = pic->data[0];

    ycbcr[1].width = info->ti.frame_width / 2;
    ycbcr[1].height = info->ti.frame_height / 2;
    ycbcr[1].stride = pic->linesize[1];
    ycbcr[1].data = pic->data[1];

    ycbcr[2].width = info->ti.frame_width / 2;
    ycbcr[2].height = info->ti.frame_height / 2;
    ycbcr[2].stride = pic->linesize[1];
    ycbcr[2].data = pic->data[2];

    th_encode_ycbcr_in(td, ycbcr);

    if(oinfo->passno==1){

        unsigned char *buffer;
        int bytes = th_encode_ctl(td, TH_ENCCTL_2PASS_OUT, &buffer, sizeof(buffer));
        if(bytes<0){
          fprintf(stderr,"Could not read two-pass data from encoder.\n");
          exit(1);
        }
        if(fwrite(buffer,1,bytes,oinfo->twopass_file)<bytes){
          fprintf(stderr,"Unable to write to two-pass data file.\n");
          exit(1);
        }
        fflush(oinfo->twopass_file);
    }

    return 0;
}

static int ff2libtheora_encode_packetout(ff2dirac_video_enc_state_t *dd, int e_o_s, ogg_packet *op)
{
    return th_encode_packetout(dd->di, e_o_s, op);
}

static int ff2libtheora_encode_init (ff2dirac ctx, ff2dirac_video_enc_state_t *dd)
{
    int i, ret;
    ff2dirac_metadata *cur_tag = ctx->metadata;
    ff2theora_privdata_t *info = av_mallocz(sizeof(ff2theora_privdata_t));
    th_enc_ctx *td;
    oggmux_info *oinfo = dd->info;

    th_info_init (&info->ti);
    // encoded size
    info->ti.frame_width = ctx->frame_width;
    info->ti.frame_height = ctx->frame_height;
    // display size
    info->ti.pic_width = ctx->picture_width;
    info->ti.pic_height = ctx->picture_height;
    info->ti.pic_x = ctx->frame_x_offset;
    info->ti.pic_y = ctx->frame_y_offset;
    info->ti.fps_numerator = ctx->framerate.num;
    info->ti.fps_denominator = ctx->framerate.den;
    /* ctx is pixel aspect ratio */
    info->ti.aspect_numerator=ctx->aspect_numerator;
    info->ti.aspect_denominator=ctx->aspect_denominator;
    info->ti.colorspace = TH_CS_UNSPECIFIED;
    if (ctx->color_primaries == AVCOL_PRI_BT470M)
        info->ti.colorspace = TH_CS_ITU_REC_470M;
    else if (ctx->color_primaries == AVCOL_PRI_BT470BG)
        info->ti.colorspace = TH_CS_ITU_REC_470BG;

    // FIXME: is all input material with fps==25 OC_CS_ITU_REC_470BG?
    // guess not, commandline option to select colorspace would be the best.
    if((ctx->fps-25)<1)
        info->ti.colorspace = TH_CS_ITU_REC_470BG;
    else if(abs(ctx->fps-30)<1)
        info->ti.colorspace = TH_CS_ITU_REC_470M;
    else

    info->ti.colorspace = TH_CS_UNSPECIFIED;

    /*Account for the Ogg page overhead.
      This is 1 byte per 255 for lacing values, plus 26 bytes per 4096 bytes for
      the page header, plus approximately 1/2 byte per packet (not accounted for
      here).*/
    info->ti.target_bitrate=(int)(64870*(ogg_int64_t)ctx->video_bitrate>>16);
    info->ti.quality = rint(ctx->video_quality*6.3);
    if (info->ti.quality < 0)
        info->ti.quality = 0;
    if (info->ti.quality > 63)
        info->ti.quality = 63;

    info->ti.keyframe_granule_shift = ilog(ctx->keyint-1);
    info->ti.pixel_fmt = TH_PF_420;

    /* no longer in the new encoder api
    info->ti.dropframes_p = 0;
    info->ti.keyframe_auto_p = 1;
    info->ti.keyframe_frequency = ctx->keyint;
    info->ti.keyframe_frequency_force = ctx->keyint;
    info->ti.keyframe_data_target_bitrate = info->ti.target_bitrate * 5;
    info->ti.keyframe_auto_threshold = 80;
    info->ti.keyframe_mindistance = 8;
    info->ti.noise_sensitivity = 1;
    // range 0-2, 0 sharp, 2 less sharp,less bandwidth
    info->ti.sharpness = ctx->sharpness;
    */

    th_comment_init(&info->tc);
    while (cur_tag) {
       if (th_comment_query(&info->tc, cur_tag->key, 0) == NULL)
           th_comment_add_tag(&info->tc, cur_tag->key, cur_tag->value);
       cur_tag = cur_tag->next;
    }

    if (oinfo->passno!=1) {
        th_comment_add_tag(&info->tc, "ENCODER", PACKAGE_STRING);
        if (strcmp(oinfo->oshash,"0000000000000000") > 0) {
            th_comment_add_tag (&info->tc, "SOURCE_OSHASH", oinfo->oshash);
        }
    }

    td = th_encode_alloc(&info->ti);
    dd->di = td;
    if (ctx->speed_level >= 0) {
        int max_speed_level;

        th_encode_ctl(td, TH_ENCCTL_GET_SPLEVEL_MAX, &max_speed_level, sizeof(int));
        if(ctx->speed_level > max_speed_level)
          ctx->speed_level = max_speed_level;
        th_encode_ctl(td, TH_ENCCTL_SET_SPLEVEL, &ctx->speed_level, sizeof(int));
    }
    /* setting just the granule shift only allows power-of-two keyframe
       spacing.  Set the actual requested spacing. */
    ret = th_encode_ctl(td, TH_ENCCTL_SET_KEYFRAME_FREQUENCY_FORCE,
                        &ctx->keyint, sizeof(ctx->keyint-1));
    if(ret<0){
        fprintf(stderr,"Could not set keyframe interval to %d.\n",(int)ctx->keyint);
    }

    if(ctx->soft_target){
        /* reverse the rate control flags to favor a 'long time' strategy */
        int arg = TH_RATECTL_CAP_UNDERFLOW;
        ret = th_encode_ctl(td, TH_ENCCTL_SET_RATE_FLAGS, &arg, sizeof(arg));
        if(ret<0)
            fprintf(stderr, "Could not set encoder flags for --soft-target\n");
        /* Default buffer control is overridden on two-pass */
        if(!oinfo->twopass && ctx->buf_delay<0){
            if((ctx->keyint*7>>1)>5*ctx->framerate.num/ctx->framerate.den)
                arg = ctx->keyint*7>>1;
            else
                arg = 5*ctx->framerate.num/ctx->framerate.den;
            ret = th_encode_ctl(td, TH_ENCCTL_SET_RATE_BUFFER, &arg,sizeof(arg));
            if(ret<0)
                fprintf(stderr, "Could not set rate control buffer for --soft-target\n");
        }
    }

    /* set up two-pass if needed */
    if(oinfo->passno==1){
      unsigned char *buffer;
      int bytes;
      bytes=th_encode_ctl(td,TH_ENCCTL_2PASS_OUT,&buffer,sizeof(buffer));
      if(bytes<0){
        fprintf(stderr,"Could not set up the first pass of two-pass mode.\n");
        fprintf(stderr,"Did you remember to specify an estimated bitrate?\n");
        exit(1);
      }
      /*Perform a seek test to ensure we can overwrite this placeholder data at
         the end; this is better than letting the user sit through a whole
         encode only to find out their pass 1 file is useless at the end.*/
      if(fseek(oinfo->twopass_file,0,SEEK_SET)<0){
        fprintf(stderr,"Unable to seek in two-pass data file.\n");
        exit(1);
      }
      if(fwrite(buffer,1,bytes,oinfo->twopass_file)<bytes){
        fprintf(stderr,"Unable to write to two-pass data file.\n");
        exit(1);
      }
      fflush(oinfo->twopass_file);
    }
    if(oinfo->passno==2){
      /* enable second pass here, actual data feeding comes later */
      if(th_encode_ctl(td,TH_ENCCTL_2PASS_IN,NULL,0)<0){
        fprintf(stderr,"Could not set up the second pass of two-pass mode.\n");
        exit(1);
      }
      if(oinfo->twopass==3){
        oinfo->videotime = 0;
        ctx->frame_count = 0;
        if(fseek(oinfo->twopass_file,0,SEEK_SET)<0){
          fprintf(stderr,"Unable to seek in two-pass data file.\n");
          exit(1);
        }
      }
    }
    if (oinfo->passno != 1 && ctx->buf_delay >= 0){
        int arg = ctx->buf_delay;
        int ret = th_encode_ctl(td, TH_ENCCTL_SET_RATE_BUFFER,
                            &ctx->buf_delay, sizeof(ctx->buf_delay));
        if (ctx->buf_delay != arg)
            fprintf(stderr, "Warning: could not set desired buffer delay of %d, using %d instead.\n", arg, ctx->buf_delay);
        if (ret < 0){
            fprintf(stderr, "Warning: could not set desired buffer delay.\n");
        }
    }

    dd->vcodec->priv_data = info;
    return 0;
}

extern double ff2libtheora_granule_time(ff2dirac_video_enc_state_t *dd, ogg_int64_t granulepos)
{
    return th_granule_time (dd->di, granulepos);
}

int ff2libtheora_granule_shift(ff2dirac_video_enc_state_t *dd)
{
    th_info *ti = &((ff2theora_privdata_t *)dd->vcodec->priv_data)->ti;
    return ti->keyframe_granule_shift;
}

int ff2libtheora_encode_comment(ff2dirac_video_enc_state_t *dd, ogg_packet *op)
{
    th_comment *tc = &((ff2theora_privdata_t *)dd->vcodec->priv_data)->tc;
    int ret = th_encode_flushheader(dd->di, tc, op);
    return ret;
}

int ff2libtheora_encode_tables(ff2dirac_video_enc_state_t *dd, ogg_packet *op)
{
    /*theora_encode_tables(dd->di, op); */
    return 0;
}

/* return frame duration in milli seconds */
ogg_int64_t ff2libtheora_get_duration(ff2dirac_video_enc_state_t *dd)
{
    ff2theora_privdata_t *tinfo = dd->vcodec->priv_data;
    ogg_int64_t duration = (tinfo->ti.fps_denominator * 1000) / tinfo->ti.fps_numerator;
    return duration;
}


int ff2libtheora_packet_iskeyframe(ogg_packet *op)
{
    return th_packet_iskeyframe(op);
}


ff2dirac_vcodec_t ff2libtheora_encoder = {
  VCODEC_THEORA,
  NULL,
  NULL,
  ff2libtheora_encode_init,
  ff2libtheora_encode_header,
  ff2libtheora_load_frame,
  ff2libtheora_encode_packetout,
  ff2libtheora_encode_clear,
  ff2libtheora_granule_time,
  ff2libtheora_granule_shift,
  ff2libtheora_encode_comment,
  ff2libtheora_encode_tables,
  ff2libtheora_get_duration,
  ff2libtheora_packet_iskeyframe
};
#endif
